from __future__ import unicode_literals

from django.urls import path

from . import views

app_name = "nautobot_deepcopy"
urlpatterns = [
    path(r"copy/<uuid:pk>/add/", views.CopyView.as_view(), name="copy"),
]



from collections import defaultdict

from django import forms
from django.db import transaction

from nautobot.extras.models import Status
from nautobot.core.forms import (
    APISelect,
    add_blank_choice,
    BootstrapMixin,
    CommentField,
    DynamicModelChoiceField,
    DynamicModelMultipleChoiceField,
)
from nautobot.dcim.choices import DeviceFaceChoices
from nautobot.dcim.models import (
    Device,
    Rack,
    Location
)
from nautobot.extras.models import Tag


class CopyForm(BootstrapMixin, forms.Form):
    name = forms.CharField(
        max_length=100,
    )
    status = DynamicModelChoiceField(
        queryset=Status.objects.get_for_model(Device),
        required=True,
    )
    location = DynamicModelChoiceField(
        queryset=Location.objects.all(),
        required=True,
    )
    rack = DynamicModelChoiceField(
        queryset=Rack.objects.all(),
        required=False,
        display_field="display_name",
        query_params={
            "location": "$location",
            "group": "$rack_group",
        },
    )
    position = forms.IntegerField(
        required=False,
        help_text="The lowest-numbered unit occupied by the device",
        widget=APISelect(
            api_url="/api/dcim/racks/{{rack}}/elevation/",
            attrs={
                "disabled-indicator": "occupied",
                "data-query-param-face": '["$face"]',
                "data-query-param-occupied": 'false',
            },
        ),
    )
    face = forms.ChoiceField(
        choices=add_blank_choice(DeviceFaceChoices),
        required=False,
        help_text="Mounted rack face",
    )
    tags = DynamicModelMultipleChoiceField(queryset=Tag.objects.all(), required=False)
    comments = CommentField()

    def __init__(self, *args, instance=None, **kwargs):
        self.instance = instance

        super().__init__(*args, **kwargs)

    def clean(self):
        # validity check if another item has the same name
        used_location = self.cleaned_data.get("location")
        used_rack = self.cleaned_data.get("rack")
        used_face = self.cleaned_data.get("face")
        used_position = self.cleaned_data.get("position")
        if Device.objects.filter(
            name=self.cleaned_data["name"], location=used_location, tenant__isnull=True
        ).exists():
            raise forms.ValidationError(
                {"name": "A device with this name already exists."}
            )

        if used_rack and used_face and used_position and Device.objects.filter(
            rack=used_rack,
            face=used_face,
            position=used_position,
        ).exists():
            raise forms.ValidationError(
                {"face": "A device with this position and face already exists."}
            )

    @transaction.atomic
    def save(self):
        self.instance.name = self.cleaned_data["name"]

        self.instance.location = self.cleaned_data["location"]
        if self.cleaned_data["status"]:
            self.instance.status = self.cleaned_data["status"]
        self.instance.comments = self.cleaned_data["comments"]
        self.instance.face = self.cleaned_data.get("face")
        self.instance.position = self.cleaned_data.get("position")
        self.instance.rack = self.cleaned_data.get("rack")

        self.instance.virtual_chassis = None
        self.instance.vc_position = None
        self.instance.asset_tag = None
        self.instance.primary_ip4_id = None
        self.instance.primary_ip6_id = None

        tags = []
        components = defaultdict(list)
        children = {}

        for tag in self.instance.tags.all():
            tags.append(tag)

        for component_type in [
            "console_ports",
            "console_server_ports",
            "power_ports",
            "power_outlets",
            "interfaces",
            "rear_ports",
            "front_ports",
            "device_bays",
        ]:
            prop = getattr(self.instance, component_type)
            for component in prop.all():
                components[component_type].append(component)

        for devicebay in self.instance.device_bays.all():
            child = devicebay.installed_device
            # only save children if available
            if child:
                child.pk = None
                child._state.adding = True
                child.name = None
                child.rack = self.cleaned_data.get("rack")
                child.primary_ip4_id = None
                child.primary_ip6_id = None
                if self.cleaned_data["status"]:
                    child.status = self.cleaned_data["status"]
                child.save()
            children[devicebay.name] = child

        self.instance.pk = None
        self.instance._state.adding = True
        self.instance.save()

        if self.cleaned_data["tags"].exists():
            for tag in self.cleaned_data["tags"].all():
                self.instance.tags.add(tag)
        else:
            for tag in tags:
                self.instance.tags.add(tag)

        for type_, component_list in components.items():
            prop = getattr(self.instance, type_)
            prop.all().delete()
            for component in component_list:
                component.device = self.instance
                component.pk = None
                component._state.adding = True
                component.cable = None
                component._cable_peer = None
                component._path = None
                if hasattr(component, "installed_device"):
                    child = children[component.name]
                    component.installed_device = child
                component.save()

        # the powerports in poweroutlets also need to be reset
        for poweroutlet in self.instance.power_outlets.all():
            if not poweroutlet.power_port:
                continue
            # TODO: will this work in all cases?
            port = [
                p
                for p in components["powerports"]
                if p.name == poweroutlet.power_port.name
            ][0]
            poweroutlet.power_port = port
            poweroutlet.save()

        return self.instance



from nautobot.extras.plugins import PluginTemplateExtension


class DeepCopyTemplate(PluginTemplateExtension):
    model = "dcim.device"

    def buttons(self):
        device = self.context["object"]

        return self.render(
            "nautobot_deepcopy/inc/buttons.html",
            {
                "device": device,
            },
        )


template_extensions = [DeepCopyTemplate]



from importlib import metadata

from nautobot.extras.plugins import PluginConfig

__version__ = metadata.version(__name__)


class DeepCopyConfig(PluginConfig):
    name = "nautobot_deepcopy"
    verbose_name = "Deep Copy"
    description = "A plugin for copying devices and their components"
    version = __version__
    author = "Gesellschaft für wissenschaftliche Datenverarbeitung mbH Göttingen"
    author_email = "netzadmin@gwdg.de"
    base_url = "nautobot_deepcopy"
    required_settings = []
    default_settings = {}
    middleware = []


config = DeepCopyConfig



from django.contrib import messages
from django.contrib.auth.mixins import PermissionRequiredMixin
from django.utils.html import escape
from django.utils.http import url_has_allowed_host_and_scheme
from django.utils.safestring import mark_safe
from django.shortcuts import redirect, render
from django.views.generic import View

from nautobot.dcim.models import Device
from nautobot.core.views.generic import GetReturnURLMixin

from .forms import CopyForm


class CopyView(PermissionRequiredMixin, GetReturnURLMixin, View):
    permission_required = "dcim.add_device"
    template_name = "nautobot_deepcopy/copy.html"

    def get(self, request, *args, pk=None, **kwargs):
        device = Device.objects.get(pk=pk)
        # Parse initial data manually to avoid setting field values as lists
        initial_data = {k: request.GET[k] for k in request.GET}
        initial_data["location"] = device.location
        initial_data["status"] = device.status
        initial_data["comments"] = device.comments
        initial_data["face"] = device.face
        initial_data["rack"] = device.rack

        form = CopyForm(initial=initial_data)

        return render(
            request,
            self.template_name,
            {
                "device": device,
                "form": form,
                "return_url": self.get_return_url(request, device),
            },
        )

    def post(self, request, *args, pk=None, **kwargs):
        device = Device.objects.get(pk=pk)
        form = CopyForm(request.POST, request.FILES, instance=device)

        if form.is_valid():
            copy = form.save()

            msg = 'Copied device <a href="{}">{}</a>'.format(
                copy.get_absolute_url(), escape(copy)
            )
            messages.success(request, mark_safe(msg))

            return_url = form.cleaned_data.get("return_url")
            if return_url is not None and url_has_allowed_host_and_scheme(
                url=return_url, allowed_hosts=request.get_host()
            ):
                return redirect(return_url)
            return redirect(self.get_return_url(request, copy))

        return render(
            request,
            self.template_name,
            {
                "device": device,
                "form": form,
                "return_url": self.get_return_url(request, device),
            },
        )
